#!/usr/bin/env python
# -*- coding: utf-8 -*-

# for backward compatibility
from torchwrench.nn.functional.new import (  # noqa: F401
    arange,
    empty,
    full,
    ones,
    rand,
    randint,
    randperm,
    zeros,
)
from torchwrench.nn.functional.transform import (  # noqa: F401
    as_tensor,
)
